/* ldap-parse-uri.h - Parse an LDAP URI.
 * Copyright (C) 2015  g10 Code GmbH
 *
 * This file is part of GnuPG.
 *
 * GnuPG is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * GnuPG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 */

#ifndef DIRMNGR_LDAP_PARSE_URI_H
#define DIRMNGR_LDAP_PARSE_URI_H

#include "../common/util.h"
#include "http.h"

extern int ldap_uri_p (const char *url);

extern gpg_error_t ldap_parse_uri (parsed_uri_t *ret_uri, const char *uri);

extern char *ldap_escape_filter (const char *filter);


#endif
