# autogen.sh configuration for GnuPG                           -*- sh -*-

display_name=GnuPG

#version_parts=3

case "$myhost:$myhostsub" in
  w32:ce)
    extraoptions="--enable-dirmngr-auto-start --disable-scdaemon "
    extraoptions="$extraoptions --disable-zip"
    ;;
  w32:)
    extraoptions="--enable-gpgtar"
    ;;
esac

case "$myhost" in
  w32)
    configure_opts="
      --with-gpg-error-prefix=@SYSROOT@
      --with-ksba-prefix=@SYSROOT@
      --with-libgcrypt-prefix=@SYSROOT@
      --with-libassuan-prefix=@SYSROOT@
      --with-zlib=@SYSROOT@
      --with-npth-prefix=@SYSROOT@
      --disable-g13
      "
    ;;

  amd64)
    configure_opts="
      --with-gpg-error-prefix=@SYSROOT@
      --with-ksba-prefix=@SYSROOT@
      --with-libgcrypt-prefix=@SYSROOT@
      --with-libassuan-prefix=@SYSROOT@
      --with-zlib=/usr/x86_64-linux-gnu/usr
      --with-pth-prefix=/usr/x86_64-linux-gnu/usr
     "
    ;;
esac


extra_aclocal_flags=""

final_info="./configure --sysconfdir=/etc --enable-maintainer-mode  && make"
