/* TinyScheme configuration.
 *
 * Copyright (C) 2016 g10 code GmbH
 *
 * This file is part of GnuPG.
 *
 * GnuPG is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * GnuPG is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <https://www.gnu.org/licenses/>.
 */

#define STANDALONE		0
#define USE_MATH		0
#define USE_CHAR_CLASSIFIERS	1
#define USE_ASCII_NAMES		1
#define USE_STRING_PORTS	1
#define USE_ERROR_HOOK		1
#define USE_TRACING		1
#define USE_COLON_HOOK		1
#define USE_DL			0
#define USE_PLIST		0
#define USE_INTERFACE		1
#define SHOW_ERROR_LINE		1
